using System;
using System.Collections.Generic;
using System.Text;
using System.Runtime.InteropServices;

namespace PCIe3021_NET
{
    public class PCIe3021
    {
        /// <summary>
        /// Store the index of the board
        /// </summary>
        private int index;

        /// <summary>
        /// Store an handle to the board. Will be used for interrupt routine for example
        /// </summary>
        private IntPtr handle;

        /// <summary>
        /// An object that will contain the timer event args
        /// Created only once, because the new command takes some time
        /// </summary>
        private TimerEventArgs timerEventArgs;

        /// <summary>
        /// An object that will contain the analog input event args
        /// Created only once, because the new command takes some time
        /// </summary>
        private AnalogInputEventArgs analogInputEventArgs;

        /// <summary>
        /// An object that will contain the end of sequence event args
        /// Created only once, because the new command takes some time
        /// </summary>
        private EndOfSequenceEventArgs endOfSequenceEventArgs;

        /// <summary>
        /// Handler to the interrupt function of the board
        /// The function pointed by this delegate will be called each time the board interrupts
        /// </summary>
        private DllImport.InterruptFunctionEventHandler interruptFunctionHandler;

        #region "Event"

        /// <summary>
        /// The timer event
        /// </summary>
        public event TimerEventHandler TimerEventReceived;

        /// <summary>
        /// Delegate to provide when an interrupt on the timer occurs
        /// </summary>
        /// <param name="source"></param>
        /// <param name="args"></param>
        public delegate void TimerEventHandler(object source, TimerEventArgs args);

        /// <summary>
        /// The analog input event
        /// </summary>
        public event AnalogInputEventHandler AnalogInputEventReceived;

        /// <summary>
        /// Delegate to provide when an interrupt on the analog input occurs
        /// </summary>
        /// <param name="source"></param>
        /// <param name="args"></param>
        public delegate void AnalogInputEventHandler(object source, AnalogInputEventArgs args);

        /// <summary>
        /// The end of sequence event
        /// </summary>
        public event EndOfSequenceEventHandler EndOfSequenceEventReceived;

        /// <summary>
        /// Delegate to provide when an end of sequence interrupt occurs
        /// </summary>
        /// <param name="source"></param>
        /// <param name="args"></param>
        public delegate void EndOfSequenceEventHandler(object source, EndOfSequenceEventArgs args);

        /// <summary>
        /// The pld fifo overflow event
        /// </summary>
        public event PldFifoOverflowEventHandler PldFifoOverflowEventReceived;

        /// <summary>
        /// Delegate to provide when a PLD fifo overflow occurs
        /// </summary>
        /// <param name="source"></param>
        public delegate void PldFifoOverflowEventHandler(object source);

        #endregion

        #region "Constructor/Destructor"

        /// <summary>
        /// Constructor of the board
        /// Cannot be called; initialize the internal variables
        /// </summary>
        private PCIe3021()
        {
            this.index = -1;
            this.handle = IntPtr.Zero;

            // Create the event args (once by board)
            this.timerEventArgs = new TimerEventArgs();
            this.analogInputEventArgs = new AnalogInputEventArgs();
            this.endOfSequenceEventArgs = new EndOfSequenceEventArgs();

            // Create the interrupt handler (once by board)
            this.interruptFunctionHandler = new DllImport.InterruptFunctionEventHandler(InterruptFunction);

            // Tell the garbage collector to keep the pointer
            GC.KeepAlive(this.interruptFunctionHandler);
        }

        /// <summary>
        /// Destructor of the board
        /// Free all the ressources
        /// </summary>
        ~PCIe3021()
        {
            DllImport.i_PCIe3021_CloseBoard(handle);
            this.index = -1;
            this.handle = IntPtr.Zero;
        }

        #endregion

        #region "Utils"

        /// <summary>
        /// Enables to get an instance of a PCIe3021 board
        /// </summary>
        /// <param name="index">The index of the board you want to open</param>
        /// <returns>null on error, instance of PCIe3021 on success</returns>
        public static PCIe3021 OpenBoard(int index)
        {
            PCIe3021 boardInstance = null;
            int nbBoard = 0;
            IntPtr handle = IntPtr.Zero;

            // Get the number of boards
            if ((nbBoard = GetBoardNumber()) == -1)
            {
                Console.WriteLine("GetBoardNumber failed");
                return null;
            }

            // Check that the index is in the range
            if (index < 0 || index > nbBoard)
            {
                Console.WriteLine("index not valid");
                return null;
            }

            // Try to open the board
            if (DllImport.i_PCIe3021_OpenBoardViaIndex((byte)index, ref handle) != 0)
            {
                Console.WriteLine("i_PCIe3021_OpenBoardViaIndex failed");
                return null;
            }

            // Create a new instance of the board
            boardInstance = new PCIe3021();

            // Set its parameters
            boardInstance.index = index;
            boardInstance.handle = handle;

            // Return the instance
            return boardInstance;
        }

        /// <summary>
        /// Enables to get the number of PCIe3021 boards plugged in the computer
        /// </summary>
        /// <returns>The number of board</returns>
        public static int GetBoardNumber()
        {
            int ret = 0;
            byte nbBoard = 0;

            // Call the DLL function
            ret = DllImport.i_PCIe3021_GetNumberOfBoards(ref nbBoard);

            // Error case
            if (ret != 0) { return -1; }

             // Success
            else { return (int)nbBoard; }
        }

        /// <summary>
        /// Enables to signal to all the listeners that a PLD fifo overflow occurs
        /// </summary>
        /// <param name="eventArgs">The arguments of the event</param>
        protected virtual void OnPldFifoOverflowEventReceived()
        {
            if (this.PldFifoOverflowEventReceived != null)
            {
                this.PldFifoOverflowEventReceived(this);
            }
        }

        #endregion

        #region "Interrupt"

        /// <summary>
        /// Interrupt function - Called each time the board generates an interrupt
        /// </summary>
        /// <param name="handle"></param>
        /// <param name="overflow"></param>
        /// <param name="source"></param>
        /// <param name="mask"></param>
        /// <param name="args"></param>
        /// <param name="argsSize"></param>
        private void InterruptFunction(IntPtr handle, byte overflow,uint source,uint mask,IntPtr args, uint argsSize )
        {
            // PLD fifo overflow ?
            if ((source & 0x10) == 0x10)
            {
                this.OnPldFifoOverflowEventReceived();
            }
            // Timer ?
            else if ((source & 2) == 2)
            {
                this.timerEventArgs.InterruptMask = mask;
                this.timerEventArgs.FifoOverflow = overflow;

                // Signal the event
                this.OnTimerEventReceived(this.timerEventArgs);
            }
            // DMA ?
            else if ((source & 4) == 4)
            {
                this.analogInputEventArgs.InterruptMask = mask;
                this.analogInputEventArgs.FifoOverflow = overflow;
                this.analogInputEventArgs.ArgsSize = argsSize;

                // Signal the event
                this.OnAnalogInputEventReceived(this.analogInputEventArgs);
            }
            // end of sequence ?
            else if ((source & 1) == 1)
            {
                byte[] data = new byte[argsSize * sizeof(ushort)];

                this.endOfSequenceEventArgs.InterruptMask = mask;
                this.endOfSequenceEventArgs.FifoOverflow = overflow;
                this.endOfSequenceEventArgs.ArgsSize = argsSize;

                // Convert the args
                Marshal.Copy(args, data, 0, (int)argsSize * sizeof(ushort));
                for (int i = 0; i < argsSize; i++)
                {
                    this.endOfSequenceEventArgs.Args[i] = BitConverter.ToUInt16(data, i * sizeof(ushort));
                }

                // Signal the event
                this.OnEndOfSequenceEventReceived(this.endOfSequenceEventArgs);
            }
        }

        /// <summary>
        /// Set the board interrupt routine
        /// </summary>
        /// <returns>0 on success, != 0 on error</returns>
        public int SetBoardIntRoutine()
        {
            return (int)DllImport.i_PCIe3021_SetBoardIntRoutine(this.handle, interruptFunctionHandler);
        }

        /// <summary>
        /// Reset the board interrupt routine
        /// </summary>
        /// <returns></returns>
        public int ResetBoardIntRoutine()
        {
            return (int)DllImport.i_PCIe3021_ResetBoardIntRoutine(this.handle);
        }

        #endregion

        #region "Digital outputs"

        /// <summary>
        /// Set the digital memory of the outputs to on
        /// </summary>
        /// <returns>0 on success, != 0 on error</returns>
        public int SetDigitalOutputMemoryOn()
        {
            return (int)DllImport.i_PCIe3021_SetDigitalOutputMemoryOn(this.handle);
        }

        /// <summary>
        /// Set the digital memory of the outputs to off
        /// </summary>
        /// <returns></returns>
        public int SetDigitalOutputMemoryOff()
        {
            return (int)DllImport.i_PCIe3021_SetDigitalOutputMemoryOff(this.handle);
        }

        /// <summary>
        /// Enables to set one channel to on (high state)
        /// </summary>
        /// <param name="channel">The index of the channel (start at 0)</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int Set1DigitalOutputOn(byte channel)
        {
            return (int)DllImport.i_PCIe3021_Set1DigitalOutputOn(this.handle, channel);
        }

        /// <summary>
        /// Enables to set one channel to off (low state)
        /// </summary>
        /// <param name="channel">The index of the channel (start at 0)</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int Set1DigitalOutputOff(byte channel)
        {
            return (int)DllImport.i_PCIe3021_Set1DigitalOutputOff(this.handle, channel);
        }

        /// <summary>
        /// Enables to set several digital outputs to on (high state)
        /// </summary>
        /// <param name="value">Digital output mask value to set (0 to 15)</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int Set4DigitalOutputsOn(byte value)
        {
            return (int)DllImport.i_PCIe3021_Set4DigitalOutputsOn(this.handle, value);
        }

        /// <summary>
        /// Enables to set several digital outputs to off (low state)
        /// </summary>
        /// <param name="value">Digital output mask value to set (0 to 15)</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int Set4DigitalOutputsOff(byte value)
        {
            return (int)DllImport.i_PCIe3021_Set4DigitalOutputsOff(this.handle, value);
        }

        /// Read 1 digital output status
        /// </summary>
        /// <param name="channel">Channel to be read (0 to 31)</param>
        /// <param name="outputStatus">Status of the digital outputs</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int Get1DigitalOutputsStatus(byte channel, ref byte outputStatus)
        {
            return (int) DllImport.i_PCIe3021_Get1DigitalOutputStatus(this.handle, channel, ref outputStatus);
        }

        /// <summary>
        /// Read 16 digital outputs status
        /// </summary>
        /// <param name="value">Status of the digital outputs</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int Get4DigitalOutputsStatus(ref byte outputStatus)
        {
            return (int)DllImport.i_PCIe3021_Get4DigitalOutputsStatus(this.handle, ref outputStatus);
        }

        #endregion

        #region "Digital inputs"

        /// <summary>
        /// Enables to read one digital
        /// </summary>
        /// <param name="channel">Channel to read (0-8)</param>
        /// <param name="inputValue">A byte that will be filled with the current status of the input (0,1)</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int Read1DigitalInput(byte channel, ref byte inputValue)
        {
            // Call the function
            return DllImport.i_PCIe3021_Read1DigitalInput(this.handle, channel, ref inputValue);
        }

        /// <summary>
        /// Enables to read 4 digital inputs (per port)
        /// </summary>
        /// <param name="inputValues">A byte that will be filled with the current status of the inputs</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int Read4DigitalInputs(ref byte inputValues)
        {
            // Call the function
            return DllImport.i_PCIe3021_Read4DigitalInputs(this.handle, ref inputValues);
        }

        #endregion

        #region "Timer"

        /// <summary>
        /// Enables to signal to all the listeners that an event on the timer occurs
        /// </summary>
        /// <param name="eventArgs">The arguments of the event</param>
        protected virtual void OnTimerEventReceived(TimerEventArgs eventArgs)
        {
            if (this.TimerEventReceived != null)
            {
                this.TimerEventReceived(this, eventArgs);
            }
        }

        /// <summary>
        /// Initialize and start the timer
        /// </summary>
        /// <param name="mode">0: TIMER_MODE_2, 1: TIMER_MODE_3</param>
        /// <param name="timeBase">0: us, 1: ms, 2: sec</param>
        /// <param name="outputMode">0: disable, 1: enabled, 2: inverted</param>
        /// <param name="interruptFlag">0: no interrupt, 1: use interrupt</param>
        /// <param name="reloadValue">maximum value of the timer</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int InitAndStartTimer(byte mode, byte timeBase,byte outputMode, byte interruptFlag, ushort reloadValue)
        {
            return DllImport.i_PCIe3021_InitAndStartTimer(this.handle, mode,timeBase,outputMode, interruptFlag, reloadValue);
        }

        /// <summary>
        /// Stop the timer
        /// </summary>
        /// <returns>0 on success, != 0 on error</returns>
        public int StopTimer()
        {
            return DllImport.i_PCIe3021_StopTimer(this.handle);
        }

        /// <summary>
        /// Read the current status of the timer
        /// </summary>
        /// <param name="status">0: stopped, 1 started</param>
        /// <param name="softwareTriggerStatus">0: no software trigger, 1 software trigger occured</param>
        /// <param name="value">current value of the timer</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int ReadTimerStatus(ref byte status, ref byte softwareTriggerStatus, ref ushort value)
        {
            return DllImport.i_PCIe3021_ReadTimerStatus(this.handle, ref status, ref softwareTriggerStatus, ref value);
        }

        /// <summary>
        /// Trigger the timer
        /// </summary>
        /// <returns>0 on success, != 0 on error</returns>
        public int TriggerTimer()
        {
            return DllImport.i_PCIe3021_TriggerTimer(this.handle);
        }

        #endregion

        #region "Analog inputs"

        /// <summary>
        /// Enables to signal to all the listeners that an event on the analog inputs occurs
        /// </summary>
        /// <param name="eventArgs">The arguments of the event</param>
        protected virtual void OnAnalogInputEventReceived(AnalogInputEventArgs eventArgs)
        {
            if (this.AnalogInputEventReceived != null)
            {
                this.AnalogInputEventReceived(this, eventArgs);
            }
        }

        /// <summary>
        /// Enables to signal to all the listeners that an end of sequence interrupt occurs
        /// </summary>
        /// <param name="eventArgs"></param>
        protected virtual void OnEndOfSequenceEventReceived(EndOfSequenceEventArgs eventArgs)
        {
            if (this.EndOfSequenceEventReceived != null)
            {
                this.EndOfSequenceEventReceived(this, eventArgs);
            }
        }

        /// <summary>
        /// Set the input mode of the board
        /// </summary>
        /// <param name="inputMode">1: single ended, 2: differential</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int SetAnalogInputMode(byte inputMode)
        {
            return DllImport.i_PCIe3021_SetAnalogInputMode(this.handle, inputMode);
        }

        /// <summary>
        /// Get the current input mode of the board
        /// </summary>
        /// <param name="inputMode">Will be filled with the current input mode. 1: single ended, 2: differential</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int GetAnalogInputMode(ref byte inputMode)
        {
            return DllImport.i_PCIe3021_GetAnalogInputMode(this.handle, ref inputMode);
        }

        /// <summary>
        /// Initialize and start a sequence acquisition
        /// </summary>
        /// <param name="sequenceCount">The number of sequence to acquire. Set to 0 if continuous acquisition</param>
        /// <param name="sequenceBeforeInterrupt">The number of sequence to be acquired per interrupt</param>
        /// <param name="channelCount">The number of channels in one sequence</param>
        /// <param name="channelArray">An array that contains the index of the channels that will be acquired</param>
        /// <param name="gainArray">An array that contains the gain of the channels</param>
        /// <param name="polarityArray">An array that contains the polarity of the channels</param>
        /// <param name="convertingTime">The converting time 5us to 65 535 000us</param>
        /// <param name="delayTime">Define the time between two sequences acquisition (5us to 65 535 000us)</param>
        /// <param name="softTriggerMode">Define the software trigger</param>
        /// <param name="extTriggerMode">Define the hardware trigger</param>
        /// <param name="sequenceTriggerCount">If trigger is enabled, and sequence mode is choosed, define the number of sequence that is acquired per trigger</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int InitAndStartAnalogInputSequence(uint sequenceCount,
            uint sequenceBeforeInterrupt,
            byte channelCount,
            byte[] channelArray,
            byte[] gainArray,
            byte[] polarityArray,
            uint convertingTime,
            uint delayTime,
            byte softTriggerMode,
            byte extTriggerMode,
            uint sequenceTriggerCount)
        {
            return DllImport.i_PCIe3021_InitAndStartAnalogInputSequence(this.handle, sequenceCount, sequenceBeforeInterrupt, channelCount,
                channelArray, gainArray, polarityArray, convertingTime, delayTime, softTriggerMode, extTriggerMode, sequenceTriggerCount);
        }

        /// <summary>
        /// Stop any analog input acquisition that is running
        /// </summary>
        /// <returns>0 on success, != 0 on error</returns>
        public int StopAnalogAcquisition()
        {
            return DllImport.i_PCIe3021_StopAnalogAcquisition(this.handle);
        }

        /// <summary>
        /// Enables to read the data from an analog input sequence
        /// </summary>
        /// <param name="size">Size of data to read</param>
        /// <param name="data">Pointer that will be filled with the read data</param>
        /// <returns> >= 0: number of data read, -1: error </returns>
        public int ReadSequenceData(uint size, ushort[] data)
        {
            return DllImport.l_PCIe3021_ReadSequenceData(this.handle, size, data);
        }

        /// <summary>
        /// Enable to read more analog inputs
        /// </summary>
        /// <param name="channelCount">The number of channels to read</param>
        /// <param name="channelArray">An array that contains the index of the channels that will be acquired</param>
        /// <param name="gainArray">An array that contains the gain of the channels</param>
        /// <param name="polarityArray">An array that contains the polarity of the channels</param>
        /// <param name="convertingTime">The converting time 5us to 65 535 000us</param>
        /// <param name="interruptFlag">0: no interrupt, 1: interrupt</param>
        /// <param name="softTriggerMode">Define the software trigger</param>
        /// <param name="extTriggerMode">Define the hardware trigger</param>
        /// <param name="value">An array that will be filled the values of the channels</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int ReadMoreAnalogInputs(byte channelCount,
            byte[] channelArray,
            byte[] gainArray,
            byte[] polarityArray,
            uint convertingTime,
            byte interruptFlag,
            byte softTriggerMode,
            byte extTriggerMode,
            ushort[] value)
        {
            return DllImport.i_PCIe3021_ReadMoreAnalogInputs(this.handle, channelCount, channelArray, gainArray, polarityArray, convertingTime, interruptFlag, softTriggerMode, extTriggerMode, value);
        }

        /// <summary>
        /// Read one analog input of the board
        /// </summary>
        /// <param name="channelIndex">The channel index of the input (0 - 15)</param>
        /// <param name="gain">The gain of the channel</param>
        /// <param name="polarity">The polarity of the channel</param>
        /// <param name="convertingTime">The converting time 5us to 65 535 000us</param>
        /// <param name="interruptFlag">0: no interrupt, 1: interrupt</param>
        /// <param name="softTriggerMode">Define the software trigger</param>
        /// <param name="extTriggerMode">Define the hardware trigger</param>
        /// <param name="value">A reference to the value that will be read</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int Read1AnalogInput(byte channelIndex,
            byte gain,
            byte polarity,
            uint convertingTime,
            byte interruptFlag,
            byte softTriggerMode,
            byte extTriggerMode,
            ref ushort value)
        {
            return DllImport.i_PCIe3021_Read1AnalogInput(this.handle, channelIndex, gain, polarity, convertingTime, interruptFlag, softTriggerMode, extTriggerMode, ref value);
        }

        /// <summary>
        /// Initialize and start an autorefresh acquisition
        /// </summary>
        /// <param name="channelCount">The number of channels to read</param>
        /// <param name="channelArray">An array that contains the index of the channels that will be acquired</param>
        /// <param name="gainArray">An array that contains the gain of the channels</param>
        /// <param name="polarityArray">n array that contains the polarity of the channels</param>
        /// <param name="convertingTime">The converting time 5us to 65 535 000us</param>
        /// <param name="value">A reference to an array that will be updated with the value of the channels</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int InitAndStartAutoRefreshAcquisition(byte channelCount,
            byte[] channelArray,
            byte[] gainArray,
            byte[] polarityArray,
            uint convertingTime,
            ref IntPtr value)
        {
            return DllImport.i_PCIe3021_InitAndStartAutoRefreshAcquisition(this.handle, channelCount, channelArray, gainArray, polarityArray, convertingTime, ref value);
        }

        /// <summary>
        /// Trigger a current analog input acquisition (software)
        /// </summary>
        /// <returns>0 on success, != 0 on error</returns>
        public int TriggerAnalogAcquisition()
        {
            return DllImport.i_PCIe3021_TriggerAnalogAcquisition(this.handle);
        }

        /// <summary>
        /// Get the number of analog inputs of the board
        /// </summary>
        /// <param name="analogInputCount">The number of analog inputs</param>
        /// <returns>0 on success, != 0 on error</returns>
        public int GetAnalogInputChannelsNbr(ref byte analogInputCount)
        {
            return DllImport.i_PCIe3021_GetAnalogInputChannelsNbr(this.handle, ref analogInputCount);
        }

        #endregion
    }
}
